from ..HitObject import HitObject
from ....Vector2 import Vector2

class RepeatPoint(HitObject):
    RepeatIndex = None
    SpanDuration = None

    def __init__(self, pos, startTime, repeatIndex, spanDuration, **kwargs):
        radius = kwargs.get('radius', None)
        super().__init__(pos, startTime, radius=radius)
        self.RepeatIndex = repeatIndex
        self.SpanDuration = spanDuration


from .HitObject import HitObject
from ...Vector2 import Vector2
from ....SliderPath import SliderPath
from .HitCircle import HitCircle
from .SliderObjects.HeadCircle import HeadCircle
from .SliderObjects.TailCircle import TailCircle
from .SliderObjects.SliderTick import SliderTick
from .SliderObjects.RepeatPoint import RepeatPoint
from functools import cmp_to_key

class Slider(HitObject):
    EndPosition = None
    EndTime = None
    Duration = None
    Path = None
    RepeatCount = None
    NestedHitObjects = []
    TickDistance = None
    LazyEndPosition = None
    LazyTravelDistance: None
    SpanDuration: None
    LegacyLastTickOffset = 36
    HeadCircle = None
    TailCircle = None
    
    Velocity = None
    SpanCount = None

    def __init__(self, pos, startTime, path, repeatCount, speedMultiplier, beatLength, mapDifficulty, radius):
        super().__init__(pos, startTime, radius=radius)
        self.Path = path
        self.EndPosition = self.Position.add(self.Path.PositionAt(1))

        self.__calculateEndTimeAndTickDistance(speedMultiplier, beatLength, mapDifficulty, repeatCount, startTime, path.expectedDistance)
        self.Duration = self.EndTime - startTime
        self.RepeatCount = repeatCount

        self.__createNestedHitObjects()
    
    def __calculateEndTimeAndTickDistance(self, speedMultiplier, beatLength, mapDifficulty, repeatCount, startTime, expectedDistance):
        scoringDistance = 100 * mapDifficulty['SliderMultiplier'] * speedMultiplier
        self.Velocity = scoringDistance / beatLength
        self.SpanCount = repeatCount + 1
        self.TickDistance = scoringDistance / mapDifficulty['SliderTickRate'] # there was a '* 1' here but honestly i dont know why would it be there
        self.EndTime = startTime + self.SpanCount * expectedDistance / self.Velocity

    def __createNestedHitObjects(self):
        self.NestedHitObjects = []

        self.__createSliderEnds()
        self.__createSliderTicks()
        self.__createRepeatPoints()

        self.NestedHitObjects = sorted(self.NestedHitObjects, key=cmp_to_key(lambda a, b: a.StartTime - b.StartTime))

        self.TailCircle.StartTime = max(self.StartTime + self.Duration / 2, self.TailCircle.StartTime - self.LegacyLastTickOffset)

    def __createSliderEnds(self):
        self.HeadCircle = HeadCircle(self.Position, self.StartTime, radius=self.Radius)
        self.TailCircle = TailCircle(self.EndPosition, self.EndTime, radius=self.Radius)

        self.NestedHitObjects.append(self.HeadCircle)
        self.NestedHitObjects.append(self.TailCircle)

    def __createSliderTicks(self):
        max_length = 100000

        length = min(max_length, self.Path.expectedDistance)
        tickDistance = min(max(self.TickDistance, 0), length)

        if(tickDistance == 0):
            return
        
        minDistanceFromEnd = self.Velocity * 10
        self.SpanDuration = self.Duration / self.SpanCount

        for span in range(self.SpanCount):
            spanStartTime = self.StartTime + span * self.SpanDuration
            Reversed = span % 2 == 1

            # for d in range(tickDistance, (length + 1), tickDistance):
            #     if(d > length - minDistanceFromEnd):
            #         break

            #     distanceProgress = d / length
            #     timeProgress = 1 - distanceProgress if Reversed else distanceProgress
            #     sliderTickPosition = self.Position.add(self.Path.PositionAt(distanceProgress))
            #     sliderTick = SliderTick(sliderTickPosition, spanStartTime + timeProgress * self.SpanDuration, span, spanStartTime, radius=self.Radius)
            #     self.NestedHitObjects.append(sliderTick)

            d = tickDistance
            while(d <= length):
                if(d > length - minDistanceFromEnd):
                    break

                distanceProgress = d / length
                timeProgress = 1 - distanceProgress if Reversed else distanceProgress
                sliderTickPosition = self.Position.add(self.Path.PositionAt(distanceProgress))
                sliderTick = SliderTick(sliderTickPosition, spanStartTime + timeProgress * self.SpanDuration, span, spanStartTime, radius=self.Radius)
                self.NestedHitObjects.append(sliderTick)

                d += tickDistance
    
    def __createRepeatPoints(self):
        for repeatIndex in range(self.RepeatCount):
            repeat = repeatIndex + 1
            repeatPosition = self.Position.add(self.Path.PositionAt(repeat % 2))
            repeatPoint = RepeatPoint(repeatPosition, self.StartTime + repeat * self.SpanDuration, repeatIndex, self.SpanDuration, radius=self.Radius)
            self.NestedHitObjects.append(repeatPoint)


from math import sqrt

class Vector2(object):
    x = y = 0.0

    def __init__(self, x, y):
        self.x = x
        self.y = y
    
    def add(self, vector):
        return Vector2(self.x + vector.x, self.y + vector.y)
    
    def subtract(self, vector):
        return Vector2(self.x - vector.x, self.y - vector.y)

    def scale(self, s_factor):
        return Vector2(self.x * s_factor, self.y * s_factor)
    
    def divide(self, div_factor):
        if(div_factor == 0):
            raise Exception("Attempt to divide vector by zero")
        return Vector2(self.x / div_factor, self.y / div_factor)

    def dot(self, vector):
        return self.x * vector.x + self.y * vector.y

    def length(self):
        return sqrt((self.x * self.x) + (self.y * self.y))
    
    def lengthSquared(self):
        return self.length() ** 2

    def distance(self, vector):
        x = self.x - vector.x
        y = self.y - vector.y
        return sqrt((x * x) + (y * y))

    def clone(self):
        return Vector2(self.x, self.y)
        
    def normalize(self):
        length = self.length()
        self.x /= length
        self.y /= length

    def almostEquals(self, vector, acceptableDifference):
        def checkNumbers(val1, val2):
            return abs(val1 - val2) <= acceptableDifference

        return checkNumbers(self.x, vector.x) and checkNumbers(self.y, vector.y)



from ..HitObject import HitObject
from ....Vector2 import Vector2

class SliderTick(HitObject):
    SpanIndex = None
    SpanStartTime = None

    def __init__(self, pos, startTime, spanIndex, spanStartTime, **kwargs):
        radius = kwargs.get('radius', None)
        super().__init__(pos, startTime, radius=radius)
        self.SpanIndex = spanIndex
        self.SpanStartTime = spanStartTime


from .HitObject import HitObject

class DifficultyHitObject(HitObject):
    TravelDistance = None
    JumpDistance = None
    Angle = None
    DeltaTime = None
    StrainTime = None

    CurrentObject = None
    LastObject = None
    LastLastObject = None

    def __init__(self, currentObject, lastObject, lastLastObject, travelDistance, jumpDistance, angle, deltaTime, strainTime):
        super().__init__(currentObject.Position, currentObject.StartTime, radius=currentObject.Radius)
        self.TravelDistance = travelDistance
        self.JumpDistance = jumpDistance
        self.Angle = angle
        self.DeltaTime = deltaTime
        self.StrainTime = strainTime

        self.CurrentObject = currentObject
        self.LastObject = lastObject
        self.LastLastObject = lastLastObject


from ..Objects.osu.HitObjects.DifficultyHitObject import DifficultyHitObject
from ..Objects.osu.HitObjects.Spinner import Spinner
from abc import ABC

class Skill(ABC):
    SINGLE_SPACING_THRESHOLD = 125
    STREAM_SPACING_THRESHOLD = 110

    def __init__(self):
        self.Previous = [] # array of DifficultyHitObject

        self.currentStrain = 1
        self.currentSectionPeak = 1
        self.strainPeaks = []

        self.SkillMultiplier = None
        self.StrainDecayBase = None

    def saveCurrentPeak(self):
        if(len(self.Previous) > 0):
            self.strainPeaks.append(self.currentSectionPeak) 

    def startNewSectionFrom(self, offset):
        if(len(self.Previous) > 0):
            self.currentSectionPeak = self.currentStrain * self.strainDecay(offset - self.Previous[0].CurrentObject.StartTime)

    def process(self, currentObject):
        self.currentStrain *= self.strainDecay(currentObject.DeltaTime)
        if(not isinstance(currentObject.CurrentObject, Spinner)):
            self.currentStrain += self.strainValueOf(currentObject) * self.SkillMultiplier

        self.currentSectionPeak = max(self.currentStrain, self.currentSectionPeak)

        self.addToHistory(currentObject)

    def difficultyValue(self):
        self.strainPeaks.sort(reverse=True)

        difficulty = 0
        weight = 1

        for strain in self.strainPeaks:
            difficulty += strain * weight
            weight *= 0.9

        return difficulty

    def strainValueOf(self, currentObject):
        pass

    def strainDecay(self, ms):
        return pow(self.StrainDecayBase, ms / 1000)

    def addToHistory(self, currentObject):
        self.Previous.insert(0, currentObject)
        if(len(self.Previous) > 2):
            self.Previous.pop()


from warnings import warn

class Logger:
    IsVerbose = False
    File = None

    def __init__(self, isVerbose, _file):
        self.IsVerbose = isVerbose
        self.File = _file

    def log(self, message, log_type):
        """Available log_types:
        "msg", "warn", "error"
        """
        # will log warnings and errors despite of verbosity
        if(log_type.lower() == 'warn'):
            warn(f'[WARN] [{self.File}] {message}', Warning)
        
        if(log_type.lower() == 'error'):
            # dont really know if it should throw an exception but oh well
            raise Exception(f'[ERROR] [{self.File}] {message}')

        if(log_type.lower() == 'msg' and self.IsVerbose):
            print(f'[MSG] [{self.File}] {message}')